@@ -1,10 +1,10 @@
-#define RUBY_VERSION "2.5.0"
-#define RUBY_RELEASE_DATE RUBY_RELEASE_YEAR_STR"-"RUBY_RELEASE_MONTH_STR"-"RUBY_RELEASE_DAY_STR
-#define RUBY_PATCHLEVEL 0
+#define RUBY_VERSION "2.5.3"
+#define RUBY_RELEASE_DATE "2018-10-18"
+#define RUBY_PATCHLEVEL 105
 
-#define RUBY_RELEASE_YEAR 2017
-#define RUBY_RELEASE_MONTH 12
-#define RUBY_RELEASE_DAY 25
+#define RUBY_RELEASE_YEAR 2018
+#define RUBY_RELEASE_MONTH 10
+#define RUBY_RELEASE_DAY 18
 
 #include "ruby/version.h"
 
